/*
 *    sparreduce.c --- reduce a chain complex with only free Abelian chain
 *                     groups as far as possible using a sequence of
 *                     elementary collapses and merging of cells.
 *                     Matrices of differentials are presented in the
 *                     PARI/GP implementation of a sparse format, and all
 *                     computations are done using sparmat library.
 *
 * Copyright (C) 2002, 2003, 2004 Alexander Shumakovitch <Shurik@Dartmouth.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program  is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see COPYING.gz. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * To load from PARI/GP:
 *    install(reduce_s_complex, "LGGGL", reduce_s_complex, "./sparreduce.so")
 */

#include <stdlib.h>
#include <limits.h>
#include <pari/pari.h>

#include "sparmat.h"

/* print some reduction statistic */
// #define PRINT_REDSTAT

/* print some debugging messages */
// #define PRINT_DEBUG

#define ERR_BAIL(msg) { ERR_MESSAGE = (msg); bailout(); }

/*
 * Type for complex sizes. 4 bytes (i.e. up to 2^32) is enough.
 */
typedef int SM_complex_t;

static SM_complex_t cplx_size = 0;			// size of the chain complex
static SM_complex_t first_group = 0;		// first non-empty chain group
static SM_complex_t last_group = 0;		// last non-empty chain group
static SparseMatrix *cplx_matrices = NULL;	// matrices of differentials
static SM_index_t *cplx_group_ranks = NULL;	// ranks of the chain groups
static SM_index_t *num_generators = NULL;		// current number of generators
static GEN pari_matrices = NULL;		// matrices in Pari format
static GEN num_entries = NULL;  		// number of matrix entries

/*
 * Free all the memory allocated in the process.
 */
static void cleanup(void)
{
	SM_complex_t i;

	if (cplx_matrices != NULL) {
		for (i = 1; i < cplx_size - 1; i++)
			kill_s_matrix(cplx_matrices + i);
		free(cplx_matrices);
	}

	if (cplx_group_ranks != NULL) free(cplx_group_ranks);
	if (num_generators != NULL) free(num_generators);

	cplx_size = 0;
	first_group = 0;
	last_group = 0;
	cplx_matrices = NULL;
	cplx_group_ranks = NULL;
	num_generators = NULL;
	pari_matrices = NULL;
	num_entries = NULL;
}

/*
 * Bailout on error.
 */
static void bailout(void)
{
	cleanup();
	err(talker, ERR_MESSAGE);
}

/*
 * Allocate memory for main arrays.
 */
static void malloc_arrays(SM_complex_t c_size)
{
	SM_complex_t i;
	char *mem_error = "malloc_arrays: not enough memory";

	cplx_size = c_size;

	cplx_matrices = (SparseMatrix *)
				malloc((cplx_size - 1) * sizeof(SparseMatrix));
	/* check for problems early, to be able to kill cplx_matrices later */
	if (cplx_matrices == NULL) ERR_BAIL(mem_error)
	for (i = 0; i < cplx_size - 1; i++) {
		cplx_matrices[i].num_rows = 0;
		cplx_matrices[i].num_cols = 0;
		cplx_matrices[i].rows = NULL;
		cplx_matrices[i].columns = NULL;
	}

	cplx_group_ranks = (SM_index_t *) malloc(cplx_size * sizeof(SM_index_t));
	num_generators = (SM_index_t *) malloc(cplx_size * sizeof(SM_index_t));
	if (cplx_group_ranks == NULL || num_generators == NULL)
		ERR_BAIL(mem_error)
}

#ifdef PRINT_DEBUG

/*
 * For testing only: print sums of squared incidence numbers for every
 * generator in the group.
 */
static void print_inums(SM_complex_t group)
{
	SM_index_t i;
	SM_value_t tmp;
	SparseVector *vec;
	SparseEntry *eptr;

	if (group < first_group || group > last_group) return;

	printf("  Group %d has %d generator(s)", group,
						num_generators[group]);
	if (group == first_group) { printf(".\n"); return; }
	else printf(": [");

 	vec = cplx_matrices[group - 1].rows;

	for (i = 0; i < cplx_group_ranks[group]; i++, vec++) {
		tmp = 0;
		for (eptr = vec -> entries; eptr != NULL; eptr = eptr->next)
			tmp += eptr->value * eptr->value;
		if (vec -> num_entries == -1) tmp = -1;
		printf("%d, ", tmp);
	}
	printf("]\n");
}

/*
 * For testing only: do print_inums for every non-empty group.
 */
static void print_all_inums(void)
{
	SM_complex_t group;

	printf("\n");
	for (group = first_group; group <= last_group; group++)
		print_inums(group);
}

#endif // #ifdef PRINT_DEBUG

/*
 * Given a matrix in PARI's sparse format, translate it into the internal one.
 * Four formats are supported:
 *   standard: the entry is [row, column, value]
 *   reduced:  the entry is [row, value * column] with value = \pm1
 *   shrunk:   the entry is value * (row * 2^32 + column) with value = \pm1
 *             row and column are assumed to be not bigger than 2^31
 *   packed:   on a 64-bit architecture the same as shrunk
 *             on a 32-bit architecture every matrix entry occupies 2 places
 *             in the VECSMALL vector: ..., row, value * column, ...
 *             row and column are assumed to be not bigger than 2^31
 */
static void assign_matrix(SparseMatrix *matr, GEN entries_list, long list_len)
{
	GEN m_entry, GEN_ptr = entries_list + 1;
	long i;
#ifdef LONG_IS_64BIT
	long tmp_val;
#endif
	SM_index_t row, column;
	SM_value_t value;
	char *matr_error = "assign_matrix: input matrix is corrupt";

	/* packed format should be treated separately */
	if (typ(entries_list) == t_VECSMALL) {
		for (i = 0; i < list_len; i++, GEN_ptr++) {
#ifdef LONG_IS_64BIT
			tmp_val = (long) *GEN_ptr;
			if (tmp_val < 0) {
				tmp_val = -tmp_val;
				value = -1;
			} else
				value = 1;
			row = tmp_val >> 32;
			column = tmp_val % ((1L << 32) - 1);
#else
			row = (SM_index_t) *(GEN_ptr++);
			column = (SM_index_t) *GEN_ptr;
			if (column < 0) {
				column = -column;
				value = -1;
			} else
				value = 1;
#endif
			if (add_m_entry(matr, row, column, value) == -1)
				bailout();
		}
		return;
	}

	for (i = 0; i < list_len; i++, GEN_ptr++) {
		if (typ(GEN_ptr) == t_VEC) {
			/* standard or reduced format */
			m_entry = (GEN) *GEN_ptr;
			if (lg(m_entry) < 3) ERR_BAIL(matr_error)

			row = (SM_index_t) itos((GEN) m_entry[1]);
			column = (SM_index_t) itos((GEN) m_entry[2]);
			if (lg(m_entry) == 4)
				/* matrix is in the standard sparse format */
				value = (SM_index_t) itos((GEN) m_entry[3]);
			else {
				/* matrix is in the reduced sparse format */
				if (column < 0) {
					column = -column;
					value = -1;
				} else
					value = 1;
			}
		} else {
			/* matrix is in the shrunk sparse format
			 *
			 * this code is strictly 32-bit only
			 * serious twicking is required to adapt it to 64 bit
			 */
			m_entry = (GEN) *GEN_ptr;
			if (typ(m_entry) != t_INT || lg(m_entry) < 4)
				ERR_BAIL(matr_error)
			value = (SM_index_t) signe(m_entry);
			row = (SM_index_t) m_entry[2];
			column = (SM_index_t) m_entry[3];
		}

		if (add_m_entry(matr, row, column, value) == -1) bailout();
	}
}

static void init_diff_matrix(SM_complex_t matrix)
{
	/* only differentials between non-empty groups are interesting */
	if (matrix < first_group || matrix >= last_group) return;

#ifdef PRINT_DEBUG
	printf("Initializing matrix number %d.", matrix);
#endif

	if (init_s_matrix(cplx_matrices + matrix, cplx_group_ranks[matrix + 1],
				cplx_group_ranks[matrix]) == -1) bailout();
	assign_matrix(cplx_matrices + matrix, (GEN) pari_matrices[matrix + 1],
					itos((GEN) num_entries[matrix + 1]));

#ifdef PRINT_DEBUG
	printf("Matrix number %d has ", matrix);
	print_s_matrix(cplx_matrices + matrix);
#endif
}

/*
 * Assign values to the main internal variables.
 */
static int init_ranks(GEN c_ranks)
{
	SM_complex_t i;

	first_group = last_group = 0;

	for (i = 0; i < cplx_size; i++) {
		num_generators[i] = cplx_group_ranks[i] =
					(SM_index_t) itos((GEN) c_ranks[i + 1]);

		if (cplx_group_ranks[i] > 0)
			last_group++;
		else {
			if (first_group == last_group) {
				first_group++;
				last_group++;
			}
		}

	}

	if (last_group > first_group) last_group--;

#ifdef PRINT_DEBUG
	printf("\n first: %d  last: %d \n", first_group, last_group);
#endif

	/* the complex is empty */
	if (first_group >= cplx_size) return -1;

	return 0;
}

/*
 * Translate a matrix in the internal format into a PARI's matrix.
 */
static GEN matr2pari(SM_complex_t group)
{
	SM_index_t i, j, row, col;
	SM_value_t val;
	SparseMatrix *matr = cplx_matrices + group;
	SparseVector *matr_cols = matr->columns, *matr_rows = matr->rows;
	SM_index_t n_rows = num_generators[group + 1];
	SM_index_t n_cols = num_generators[group];
	SM_index_t n_m_rows = matr->num_rows, n_m_cols = matr->num_cols;
	GEN pari_matr = cgetg(n_cols + 1, t_MAT), pari_vec;
	char *matr_error = "matr2pari: matrix is corrupt";

	for (i = 1, col = 1; i <= n_cols; i++, col++) {
		while(matr_cols[col - 1].num_entries == -1 && col <= n_m_cols)
			col++;
		if (col > n_m_cols) ERR_BAIL(matr_error)

		pari_vec = cgetg(n_rows + 1, t_COL);
		for (j = 1, row = 1; j <= n_rows; j++, row++) {
			while(matr_rows[row - 1].num_entries == -1 &&
					row <= n_m_rows) row++;
			if (row > n_m_rows) ERR_BAIL(matr_error)

			/* remove_m_entry checks much more than get_m_entry */
			if ((val = remove_m_entry(matr, row, col)) == ERR_MVAL)
				bailout();

			pari_vec[j] = lstoi(val);
		}
		pari_matr[i] = (long) pari_vec;
	}

	return pari_matr;
}

/*
 * Prepare the result to be sent back to PARI.
 */
static GEN feed2pari()
{
	SM_complex_t i, group;
	GEN main_vec = cgetg(3, t_VEC);
	GEN matrices_vec = cgetg(cplx_size, t_VEC);
	GEN numgen_vec = cgetg(cplx_size + 1, t_VEC);

	for (i = 1; i <= cplx_size; i++) numgen_vec[i] = zero;
	main_vec[1] = (long) numgen_vec;

	for (i = 1; i < cplx_size; i++) matrices_vec[i] = zero;
	main_vec[2] = (long) matrices_vec;

	if (first_group >= cplx_size) return main_vec;

	for (group = first_group; group <= last_group; group++) {
		/* nothing to do with an empty group */
		if (num_generators[group] == 0) continue;
		numgen_vec[group + 1] = lstoi(num_generators[group]);

		/* no matrices after the last group */
		if (group == last_group) continue;

		/* no matrices with zero size */
		if (num_generators[group + 1] == 0) continue;

		matrices_vec[group + 1] = (long) matr2pari(group);
	}

	/*
	for (group = first_group; group < last_group; group++)
		if (!is_matr_empty(cplx_matrices + group))
			ERR_BAIL("feed2pari: not all matrices are returned")
	*/

	return main_vec;
}

/*
 * Kill a generator gen_num in the group.
 */
static void kill_gen(SM_complex_t group, SM_index_t gen_num)
{
	if (group > first_group)
		if (erase_m_row(cplx_matrices + group - 1, gen_num, 1) == -1)
			bailout();

	if (group < last_group)
		if (erase_m_column(cplx_matrices + group, gen_num, 1) == -1)
			bailout();

	num_generators[group]--;
}

/*
 * Subtract two generators gens[0] and gens[1] in the group from each other
 * with a sign gen_sgn. If rdir < 0, use the dual complex.
 */
static void sub_gens(SM_complex_t group,
		SM_index_t gens[2], int gen_sgn, int rdir)
{
	SM_index_t gen1 = gens[0], gen2 = gens[1];
	SM_value_t tmp;

	if (rdir > 0 && group < last_group) {
		tmp = add_m_cols(cplx_matrices + group, gen1, gen2, -gen_sgn);
		if (tmp == -1) bailout();
	}
	if (rdir < 0 && group > first_group) {
		tmp = add_m_rows(cplx_matrices + group - 1,
							gen1, gen2, -gen_sgn);
		if (tmp == -1) bailout();
	}

	kill_gen(group, gen2);
}

/*
 * Find a (unique) generator in group-1 which is incident to gen in the group
 * with incidence number \pm1. Return -1 if the incidence number is not \pm1.
 * If rdir < 0, use the dual complex.
 */
static SM_index_t find_inc_gen(SM_complex_t group, SM_index_t gen, int rdir)
{
	SparseEntry *entry;

	entry = rdir > 0 ? cplx_matrices[group - 1].rows[gen - 1].entries:
			   cplx_matrices[group].columns[gen - 1].entries;

	if (ABSFUNC(entry->value) == 1) return entry->index;
	else return -1;
}

/*
 * Find two generators in group-1 which are incident to gen in the group
 * with incidence numbers \pm1. Return the difference in signs as well.
 * Return ERR_MVAL if the incidence numbers are not \pm1.
 * If rdir < 0, use the dual complex.
 */
static SM_value_t find_2_inc_gens(SM_complex_t group, SM_index_t gen,
						SM_index_t inc_gens[2], int rdir)
{
	SparseEntry *entry1, *entry2;

	entry1 = rdir > 0 ? cplx_matrices[group - 1].rows[gen - 1].entries:
			    cplx_matrices[group].columns[gen - 1].entries;
	entry2 = entry1->next;


	if (ABSFUNC(entry1->value) == 1 && ABSFUNC(entry2->value) == 1) {
		inc_gens[0] = entry1->index;
		inc_gens[1] = entry2->index;
		return entry1->value * entry2->value;
	} else return ERR_MVAL;
}

/*
 * Reduce as many generators as possible in a given group.
 * Return 1 if some reduction was done and 0 otherwise.
 * If rdir < 0, do the dual complex reduction.
 */
static int reduce_gens(SM_complex_t group, int rdir)
{
	SM_index_t i, cnt1 = 0, cnt2 = 0, gen, inc_gen, inc_gens[2];
	SM_value_t r_sgn;
	int isfound = 0;

	SparseVector *inum_vectors;
	char *gen_error = "reduce_gens: generator is not killed cleanly";

	/* check that the differential matrices are already initialized */
	if (cplx_matrices[group - 1].rows == 0) init_diff_matrix(group - 1);
	if (cplx_matrices[group].rows == 0) init_diff_matrix(group);

	gen = 1;
	inum_vectors = rdir > 0 ? cplx_matrices[group - 1].rows:
				  cplx_matrices[group].columns;

	/* in the dual complex the reduction should be done backwards,
	 * otherwise too much is left (no ideas why) */
	if (rdir < 0) {
		gen = cplx_group_ranks[group];
		inum_vectors += cplx_group_ranks[group] - 1;
	}

	for (i = 1; i <= cplx_group_ranks[group]; i++,
					gen += rdir, inum_vectors += rdir) {
		switch (inum_vectors->num_entries) {
			case 1:
				/* free wall is found */
				inc_gen = find_inc_gen(group, gen, rdir);
				if (inc_gen != -1) {
					cnt1++;
					isfound = 1;
					kill_gen(group - rdir, inc_gen);
					if (inum_vectors->num_entries != 0)
						ERR_BAIL(gen_error);
					kill_gen(group, gen);
				}
				break;
			case 2:
				/* a wall between two rooms is found */
				r_sgn = find_2_inc_gens(group,
							gen, inc_gens, rdir);
				if (r_sgn != ERR_MVAL) {
					cnt2++;
					isfound = 1;
					sub_gens(group - rdir,
							inc_gens, r_sgn, rdir);
					if (inum_vectors->num_entries != 0)
						ERR_BAIL(gen_error);
					kill_gen(group, gen);
				}
				break;
		}
	}

#ifdef PRINT_REDSTAT
	if (isfound) printf("%d %d | ", cnt1, cnt2);
#endif

	return isfound;
}

/*
 * Reduce a chain complex with only free Abelian chain groups as far as
 * possible using a sequence of elementary collapses and merging of cell.
 *
 * Arguments are:
 *   size (length) of the complex probably without zero groups at the ends
 *   ranks of the chain groups
 *   matrices of chain differentials in the sparse format
 *   lengths of arrays representing the matrices
 *   direction of the reduction: 0 ==> no reduction; 1 ==> from left to right;
 *     2 ==> from right to left in the dual complex; 3 = 1 + 2.
 *
 * The return value is a 2-component vector which contain
 *   ranks of the chain groups after the reduction
 *   matrices of chain differentials after the reduction
 *     (matrices of size 0 are substituted with 0 for better visualization)
 */
GEN reduce_s_complex(long c_size, GEN c_ranks, GEN d_matrices,
		GEN matr_lengths, long direction)
{
	GEN answer;
	SM_complex_t group;
	int cnt;

	malloc_arrays((SM_complex_t) c_size);
	pari_matrices = d_matrices;
	num_entries = matr_lengths;

	if (init_ranks(c_ranks) == -1) {
		/* complex is empty */
		answer = feed2pari();
		cleanup();
		return answer;
	}

	/* try to go from left to right first */
	if (direction & 1) {
#ifdef PRINT_REDSTAT
		printf("\n   ");
#endif
		for (group = first_group + 1; group <= last_group; group++) {
			cnt = 0;
#ifdef PRINT_REDSTAT
			printf("%d: ", group);
#endif
			/* try to reduce generators in this group
			 * repeat the procedure if something was reduced */
			while (reduce_gens(group, 1)) cnt++;
#ifdef PRINT_REDSTAT
			printf("%d;  ", cnt);
#endif

			/* stop if we (apparently) reached non-zero homology
			 * and still have to go in another direction */
		//	if ((direction & 2) && num_generators[group - 1] > 0)
		//		break;
		}
	}

	/* and now from right to left */
	if (direction & 2) {
#ifdef PRINT_REDSTAT
		printf("\n   ");
#endif
		for (group = last_group - 1; group >= first_group; group--) {
			cnt = 0;
#ifdef PRINT_REDSTAT
			printf("%d: ", group);
#endif
			/* try to reduce generators in this group
			 * repeat the procedure if something was reduced */
			while (reduce_gens(group, -1)) cnt++;
#ifdef PRINT_REDSTAT
			printf("%d;  ", cnt);
#endif
		}
	}

	answer = feed2pari();
	cleanup();
	return answer;
}
